/*
 * Part of "LIAB - Linux in a Box"
 * Copyright 2013 Martin von Gagern
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/inotify.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "watchdog.h"

static long startlog;
static int inot;
static pid_t child;

static sig_atomic_t child_dead = 0, writing_log = 0;

void reap(int sig) {
  int status;
  CHK(waitpid(child, &status, WNOHANG), -1);
  child_dead = 1;
  if (!writing_log)
    exit(EXIT_SUCCESS);
}

void copylog() {
  char logcopy[1024];
  time_t t;
  FILE *outf;
  int ch;

  time(&t);
  strftime(logcopy, sizeof(logcopy),
	   "/var/log/persist/wd_%F_%H-%M-%S%z.log", localtime(&t));
  outf = fopen(logcopy, "wt");
  if (outf == NULL) {
    syslog(LOG_ERR, "Failed to open '%s': %s", logcopy, strerror(errno));
    return;
  }
  syslog(LOG_NOTICE, "Copying log to '%s'", logcopy);
  sleep(1);
  if (fseek(yyin, startlog, SEEK_SET) == -1) return;
  while ((ch = getc(yyin)) != EOF)
    if (putc(ch, outf) == EOF)
      return;
  fsync(fileno(outf));
  fclose(outf);
}

void restart() {
  writing_log = 1;
  CHK(kill(child, SIGTERM), -1);
  copylog();
  writing_log = 0;
  if (!child_dead)
    sleep(10); /* give child some time to die */
  exit(EXIT_SUCCESS);
}

void fix_input() {
  char event[sizeof(struct inotify_event) + NAME_MAX + 1];
  CHK(!ferror(yyin), 0);
  CHK(read(inot, event, sizeof(event)), 0);
  fseek(yyin, 0, SEEK_CUR);
}

int main(int argc, char** argv) {
  char* fname;
  char** args;
  int wd;
  int i;

  if (argc < 3) {
    fprintf(stderr, "Usage: %s file program [args...]\n", argv[0]);
    return EXIT_FAILURE;
  }

  openlog("watchdog", 0, LOG_USER);

  fname = argv[1];
  argv[1] = argv[2];
  CHK(yyin = fopen(fname, "rt"), NULL);
  CHK(fseek(yyin, 0, SEEK_END), -1);
  CHK(startlog = ftell(yyin), -1);
  CHK(inot = inotify_init(), -1);
  CHK(wd = inotify_add_watch(inot, fname, IN_MODIFY), -1);

  CHK(signal(SIGCHLD, reap), SIG_ERR);
  CHK(args = malloc((argc - 1)*sizeof(char*)), NULL);
  for (i = 2; i != argc; ++i)
    args[i - 2] = argv[i];
  args[argc - 2] = NULL;
  CHK(child = fork(), -1);
  if (child == 0)
    CHK(execvp(args[0], args), -1);

  yylex();
  /* should not return here. */
  return EXIT_SUCCESS;
}
